# copyright ############################### #
# This file is part of the Xdeps Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

from .tasks import Manager
from . import madxutils
from .table import Table

from .optimize import (Optimize, Vary, Target, TargetList, VaryList,
                        TargetInequality, Action)

from ._version import __version__



import numpy as np
from ..general import _print

from scipy.optimize import fsolve, minimize
from .jacobian import JacobianSolver


class Vary:
    def __init__(self, name, container, limits=None, step=None, weight=None):

        if weight is None:
            weight = 1.

        if limits is None:
            limits = (-1e200, 1e200)
        else:
            assert len(limits) == 2, '`limits` must have length 2.'

        if step is None:
            step = 1e-10

        assert weight > 0, '`weight` must be positive.'

        self.name = name
        self.limits = np.array(limits)
        self.step = step
        self.weight = weight
        self.container = container

class VaryList:
    def __init__(self, vars, container, **kwargs):
        self.vary_objects = [Vary(vv, container, **kwargs) for vv in vars]

class Target:
    def __init__(self, tar, value, tol=None, weight=None, scale=None,
                 action=None):

        if scale is not None and weight is not None:
            raise ValueError(
                'Cannot specify both `weight` and `scale` for a target.')

        if scale is not None:
            weight = scale

        self.tar = tar
        self.action = action
        self.value = value
        self.tol = tol
        self.weight = weight
        self._at_index = None

    @property
    def scale(self):
        return self.weight

    @scale.setter
    def scale(self, value):
        self.weight = value

    def eval(self, data):
        res = data[self.action]
        if callable(self.tar):
            return self.tar(res)
        else:
            return res[self.tar]

class TargetList:
    def __init__(self, tars, **kwargs):
        self.targets = [Target(tt, **kwargs) for tt in tars]


class TargetInequality(Target):

    def __init__(self, tar, ineq_sign, rhs, tol=None, scale=None):
        super().__init__(tar, value=0, tol=tol, scale=scale)
        assert ineq_sign in ['<', '>'], ('ineq_sign must be either "<" or ">"')
        self.ineq_sign = ineq_sign
        self.rhs = rhs

    def eval(self, tw):
        val = super().eval(tw)
        if self.ineq_sign == '<' and val < self.rhs:
            return 0
        elif self.ineq_sign == '>' and val > self.rhs:
            return 0
        else:
            return val - self.rhs

class Action:
    def prepare(self):
        pass
    def run(self):
        return dict()

class MeritFunctionForMatch:

    def __init__(self, vary, targets, actions, return_scalar,
                 call_counter, verbose, tw_kwargs, steps_for_jacobian):

        self.vary = vary
        self.targets = targets
        self.actions = actions
        self.return_scalar = return_scalar
        self.call_counter = call_counter
        self.verbose = verbose
        self.tw_kwargs = tw_kwargs
        self.steps_for_jacobian = steps_for_jacobian
        self.found_point_within_tolerance = False

    def _x_to_knobs(self, x):
        knob_values = np.array(x).copy()
        for ii, vv in enumerate(self.vary):
            if vv.weight is not None:
                knob_values[ii] *= vv.weight
        return knob_values

    def _knobs_to_x(self, knob_values):
        x = np.array(knob_values).copy()
        for ii, vv in enumerate(self.vary):
            if vv.weight is not None:
                x[ii] /= vv.weight
        return x

    def __call__(self, x):

        _print(f"Matching: model call n. {self.call_counter}       ",
                end='\r', flush=True)
        self.call_counter += 1

        knob_values = self._x_to_knobs(x)

        for vv, val in zip(self.vary, knob_values):
            vv.container[vv.name] = val

        if self.verbose:
            _print(f'x = {knob_values}')

        res_data = {}
        failed = False
        for aa in self.actions:
            res_data[aa] = aa.run()
            if res_data[aa] == 'failed':
                failed = True
                break

        if failed:
            err_values = [1e100 for tt in self.targets]
        else:
            res_values = []
            target_values = []
            for tt in self.targets:
                res_values.append(tt.eval(res_data))
                target_values.append(tt.value)
            self._last_data = res_data # for debugging

            res_values = np.array(res_values)
            target_values = np.array(target_values)
            err_values = res_values - target_values

            if self.verbose:
                _print(f'   f(x) = {res_values}')

            tols = 0 * err_values
            for ii, tt in enumerate(self.targets):
                tols[ii] = tt.tol

            if self.verbose:
                _print(f'   err/tols = {err_values/tols}')

            if np.all(np.abs(err_values) < tols):
                err_values *= 0
                self.found_point_within_tolerance = True
                if self.verbose:
                    _print('Found point within tolerance!')

            for ii, tt in enumerate(self.targets):
                if tt.weight is not None:
                    err_values[ii] *= tt.weight

        if self.return_scalar:
            return np.sum(err_values * err_values)
        else:
            return np.array(err_values)

    def get_jacobian(self, x):
        x = np.array(x).copy()
        steps = self._knobs_to_x(self.steps_for_jacobian)
        assert len(x) == len(steps)
        f0 = self(x)
        if np.isscalar(f0):
            jac = np.zeros((1, len(x)))
        else:
            jac = np.zeros((len(f0), len(x)))
        for ii in range(len(x)):
            x[ii] += steps[ii]
            jac[:, ii] = (self(x) - f0) / steps[ii]
            x[ii] -= steps[ii]
        return jac

class Optimize:

    def __init__(self, vary, targets, restore_if_fail=True,
                 solver=None,
                 verbose=False, assert_within_tol=True,
                 solver_options={}, **kwargs):

        if isinstance(vary, (str, Vary)):
            vary = [vary]

        input_vary = vary
        vary = []
        for ii, rr in enumerate(input_vary):
            if isinstance(rr, Vary):
                vary.append(rr)
            elif isinstance(rr, str):
                vary.append(Vary(rr))
            elif isinstance(rr, (list, tuple)):
                raise ValueError('Not supported')
            elif isinstance(rr, VaryList):
                vary += rr.vary_objects
            else:
                raise ValueError(f'Invalid vary setting {rr}')

        input_targets = targets
        targets = []
        for ii, tt in enumerate(input_targets):
            if isinstance(tt, Target):
                targets.append(tt)
            elif isinstance(tt, (list, tuple)):
                targets.append(Target(*tt))
            elif isinstance(tt, TargetList):
                targets += tt.targets
            else:
                raise ValueError(f'Invalid target element {tt}')

        actions = []
        for tt in targets:
            if tt.weight is None:
                tt.weight = 1.
            if tt.weight <= 0:
                raise ValueError('`weight` must be positive.')

            if tt.action not in actions:
                actions.append(tt.action)

        for aa in actions:
            aa.prepare()

        data0 = {}
        for aa in actions:
            data0[aa] = aa.run()
            assert data0[aa] != 'failed', (
                f'Action {aa} failed to compute initial data.')

        for tt in targets:
            if tt.value == 'preserve':
                tt.value = tt.eval(data0)

        if solver is None:
            solver = 'jacobian'

        if verbose:
            _print(f'Using solver {solver}')

        steps = []
        knob_limits = []
        for vv in vary:
            steps.append(vv.step)
            knob_limits.append(vv.limits)

        assert solver in ['fsolve', 'bfgs', 'jacobian'], (
                        f'Invalid solver {solver}.')

        return_scalar = {'fsolve': False, 'bfgs': True, 'jacobian': False}[solver]

        _err = MeritFunctionForMatch(
                    vary=vary, targets=targets,
                    actions=actions,
                    return_scalar=return_scalar, call_counter=0, verbose=verbose,
                    tw_kwargs=kwargs, steps_for_jacobian=steps)

        knob_limits = np.array(knob_limits)
        x_lim_low = _err._knobs_to_x(np.atleast_1d(np.squeeze(knob_limits[:, 0])))
        x_lim_high = _err._knobs_to_x(np.atleast_1d(np.squeeze(knob_limits[:, 1])))
        x_limits = [(hh, ll) for hh, ll in zip(x_lim_low, x_lim_high)]

        _jac= _err.get_jacobian

        self._err = _err
        self._jac = _jac
        self.solver = solver
        self.x_limits = x_limits
        self.assert_within_tol = assert_within_tol
        self.verbose = verbose
        self.restore_if_fail = restore_if_fail
        self.solver_options = solver_options

    @property
    def vary(self):
        return self._err.vary

    @property
    def targets(self):
        return self._err.targets

    def _extract_knob_values(self):
        res = []
        for vv in self.vary:
            val = vv.container[vv.name]
            if hasattr(val, '_value'):
                res.append(val._value)
            else:
                res.append(val)
        return res

    def solve(self, x0=None):

        x0 = self._err._knobs_to_x(self._extract_knob_values())
        try:
            if self.solver == 'jacobian':
                jac_solver = JacobianSolver(
                    func=self._err, limits=self.x_limits, verbose=self.verbose,
                    **self.solver_options)
                res = jac_solver.solve(x0=x0.copy())
                result_info = {'jac_solver': jac_solver, 'res': res}
            elif self.solver == 'fsolve':
                raise NotImplementedError # Untested
                (res, infodict, ier, mesg) = fsolve(self._err, x0=x0.copy(),
                    full_output=True, fprime=self._jac)
                if ier != 1:
                    raise RuntimeError("fsolve failed: %s" % mesg)
                result_info = {
                    'res': res, 'info': infodict, 'ier': ier, 'mesg': mesg}
            elif self.solver == 'bfgs':
                raise NotImplementedError # Untested
                optimize_result = minimize(self._err, x0=x0.copy(), method='L-BFGS-B',
                            bounds=self.x_limits,
                            jac=self._jac, options={'gtol':0})
                result_info = {'optimize_result': optimize_result}
                res = optimize_result.x


            if self.assert_within_tol and not self._err.found_point_within_tolerance:
                raise RuntimeError('Could not find point within tolerance.')

            for vv, rr in zip(self.vary, self._err._x_to_knobs(res)):
                vv.container[vv.name] = rr
        except Exception as err:
            if self.restore_if_fail:
                knob_values0 = self._err._x_to_knobs(x0)
                for ii, vv in enumerate(self.vary):
                    vv.container[vv.name] = knob_values0[ii]
            _print('\n')
            raise err
        _print('\n')
        return result_info




# copyright ############################### #
# This file is part of the Xdeps Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #


from dataclasses import dataclass, field
import operator, builtins, math

_binops = {
    "+": operator.add,
    "-": operator.sub,
    "*": operator.mul,
    "@": operator.matmul,
    "/": operator.truediv,
    "//": operator.floordiv,
    "%": operator.mod,
    "**": operator.pow,
    "&": operator.and_,
    "|": operator.or_,
    "^": operator.xor,
    "<": operator.lt,
    "<=": operator.le,
    "==": operator.eq,
    "!=": operator.ne,
    ">=": operator.ge,
    ">": operator.gt,
    ">>": operator.rshift,
    "<<": operator.lshift,
}

_mutops = {
    "+": operator.iadd,
    "//": operator.ifloordiv,
    "<<": operator.ilshift,
    "@": operator.imatmul,
    "%": operator.imod,
    "*": operator.imul,
    "**": operator.ipow,
    ">>": operator.irshift,
    "-": operator.isub,
    "/": operator.itruediv,
    "^": operator.ixor,
}


_unops = {"-": operator.neg, "+": operator.pos, "~": operator.invert}

_builtins = {
    "divmod": builtins.divmod,
    "abs": builtins.abs,
    "complex": builtins.complex,
    "int": builtins.int,
    "float": builtins.float,
    "round": builtins.round,
    "trunc": math.trunc,
    "floor": math.floor,
    "ceil": math.ceil,
}


def _pr_binop():
    for sy, op in _binops.items():
        fn = op.__name__.replace("_", "")
        rr = fn.capitalize()
        fmt = f"""
       def __{fn}__(self, other):
           return {rr}Ref(self, other)  # type: ignore

       def __r{fn}__(self, other):
           return {rr}Ref(other, self)  # type: ignore"""
        print(fmt)


def _pr_builtins():
    for sy, op in _builtins.items():
        fn = op.__name__.replace("_", "")
        rr = fn.capitalize()
        fmt = f"""
       def __{fn}__(self, other): 
           return {rr}Ref(self, other)  # type: ignore"""
        print(fmt)

    ### careful with abs, round, floor, ceil


def _pr_mutops():
    for sy, op in _mutops.items():
        fn = op.__name__.replace("_", "")
        fmt = f"""
    def __{fn}__(self, other):
        newexpr=self._expr
        if newexpr:
            return newexpr{sy}other
        else:
            return self._get_value(){sy}other"""
        print(fmt)


class ARef:
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        raise ValueError("Cannot instantiate ARef")

    @staticmethod
    def _mk_value(value):
        if isinstance(value, ARef):
            return value._get_value()
        else:
            return value

    @property
    def _value(self):
        return self._get_value()

    # order of precedence
    def __call__(self, *args, **kwargs):
        return CallRef(self, args, kwargs)

    def __getitem__(self, item):
        return ItemRef(self, item, self._manager)

    def __getattr__(self, attr):
        if attr.startswith("__array_"):  # numpy crashes without
            # print(self,attr)
            raise AttributeError
        if attr == '_manager':
            raise RuntimeError("This should not happen...")
        return AttrRef(self, attr, self._manager)

    # Untested:

    # def __getstate__(self):
    #     raise ValueError("Cannot pickle ARef")

    # def __setstate__(self, state):
    #     raise ValueError("Cannot pickle ARef")

    # numerical unary  operator
    def __neg__(self):
        return NegRef(self)  # type: ignore

    def __pos__(self):
        return PosRef(self)  # type: ignore

    def __invert__(self):
        return InvertRef(self)  # type: ignore

    # numerical binary operators

    def __add__(self, other):
        return AddRef(self, other)  # type: ignore

    def __radd__(self, other):
        return AddRef(other, self)  # type: ignore

    def __sub__(self, other):
        return SubRef(self, other)  # type: ignore

    def __rsub__(self, other):
        return SubRef(other, self)  # type: ignore

    def __mul__(self, other):
        return MulRef(self, other)  # type: ignore

    def __rmul__(self, other):
        return MulRef(other, self)  # type: ignore

    def __matmul__(self, other):
        return MatmulRef(self, other)  # type: ignore

    def __rmatmul__(self, other):
        return MatmulRef(other, self)  # type: ignore

    def __truediv__(self, other):
        return TruedivRef(self, other)  # type: ignore

    def __rtruediv__(self, other):
        return TruedivRef(other, self)  # type: ignore

    def __floordiv__(self, other):
        return FloordivRef(self, other)  # type: ignore

    def __rfloordiv__(self, other):
        return FloordivRef(other, self)  # type: ignore

    def __mod__(self, other):
        return ModRef(self, other)  # type: ignore

    def __rmod__(self, other):
        return ModRef(other, self)  # type: ignore

    def __pow__(self, other):
        return PowRef(self, other)  # type: ignore

    def __rpow__(self, other):
        return PowRef(other, self)  # type: ignore

    def __and__(self, other):
        return AndRef(self, other)  # type: ignore

    def __rand__(self, other):
        return AndRef(other, self)  # type: ignore

    def __or__(self, other):
        return OrRef(self, other)  # type: ignore

    def __ror__(self, other):
        return OrRef(other, self)  # type: ignore

    def __xor__(self, other):
        return XorRef(self, other)  # type: ignore

    def __rxor__(self, other):
        return XorRef(other, self)  # type: ignore

    def __lt__(self, other):
        return LtRef(self, other)  # type: ignore

    def __rlt__(self, other):
        return LtRef(other, self)  # type: ignore

    def __le__(self, other):
        return LeRef(self, other)  # type: ignore

    def __rle__(self, other):
        return LeRef(other, self)  # type: ignore

    def __eq__(self, other):
        return EqRef(self, other)  # type: ignore

    def __req__(self, other):
        return EqRef(other, self)  # type: ignore

    def __ne__(self, other):
        return NeRef(self, other)  # type: ignore

    def __rne__(self, other):
        return NeRef(other, self)  # type: ignore

    def __ge__(self, other):
        return GeRef(self, other)  # type: ignore

    def __rge__(self, other):
        return GeRef(other, self)  # type: ignore

    def __gt__(self, other):
        return GtRef(self, other)  # type: ignore

    def __rgt__(self, other):
        return GtRef(other, self)  # type: ignore

    def __rshift__(self, other):
        return RshiftRef(self, other)  # type: ignore

    def __rrshift__(self, other):
        return RshiftRef(other, self)  # type: ignore

    def __lshift__(self, other):
        return LshiftRef(self, other)  # type: ignore

    def __rlshift__(self, other):
        return LshiftRef(other, self)  # type: ignore

    def __divmod__(self, other):
        return BDivmodRef(self, other)  # type: ignore

    def __pow__(self, other):
        return PowRef(self, other)  # type: ignore

    def __round__(self, other=0):
        return BRoundRef(self, other)  # type: ignore

    def __trunc__(self, other):
        return BTruncRef(self, other)  # type: ignore

    def __floor__(self, other):
        return BFloorRef(self, other)  # type: ignore

    def __ceil__(self, other):
        return BCeilRef(self, other)  # type: ignore

    def __abs__(self):
        return BAbsRef(self)  # type: ignore

    def __complex__(self):
        return BComplexRef(self)  # type: ignore

    def __int__(self):
        return BIntRef(self)  # type: ignore

    def __float__(self):
        return BFloatRef(self)  # type: ignore


class MutableRef(ARef):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        raise ValueError("Cannot instantiate MutableRef")

    def __setitem__(self, key, value):
        ref = ItemRef(self, key, self._manager)
        self._manager.set_value(ref, value)

    def __setattr__(self, attr, value):
        if attr[0] == "_" and attr in ["_expr", "_exec"]:
            raise ValueError(f"`{attr}` is a special keyword and cannot be assigned.")
        ref = AttrRef(self, attr, self._manager)
        self._manager.set_value(ref, value)

    def _eval(self, expr, gbl=None):
        if gbl is None:
            gbl = {}
        return eval(expr, gbl, self)

    def _exec(self, expr, gbl=None):
        if gbl is None:
            gbl = {}
        exec(expr, gbl, self)

    @property
    def _tasks(self):
        return self._manager.tartasks[self]

    def _find_dependant_targets(self):
        return self._manager.find_deps([self])

    @property
    def _expr(self):
        if self in self._manager.tasks:
            task = self._manager.tasks[self]
            if hasattr(task, "expr"):
                return task.expr

    def _info(self, limit=10):
        print(f"#  {self}._get_value()")
        try:
            value = self._get_value()
            print(f"   {self} = {value}")
        except:
            print(f"#  {self} has no value")
        print()

        if self in self._manager.tasks:
            task = self._manager.tasks[self]
            print(f"#  {self}._expr")
            print(f"   {task}")
            print()
            if hasattr(task, "info"):
                task.info()
        else:
            print(f"#  {self}._expr is None")
            print()

        refs = self._manager.find_deps([self])[1:]
        if len(refs) == 0:
            print(f"#  {self} does not influence any target")
            print()
        else:
            print(f"#  {self}._find_dependant_targets()")
            for tt in refs[:limit]:
                if tt._expr is not None:
                    print(f"   {tt}")
            if len(refs) > limit:
                print(f"   ... set _info(limit={len(refs)}) to get all lines")
            print()

    def __iadd__(self, other):
        newexpr = self._expr
        if newexpr:
            return newexpr + other
        else:
            return self._get_value() + other

    def __ifloordiv__(self, other):
        newexpr = self._expr
        if newexpr:
            return newexpr // other
        else:
            return self._get_value() // other

    def __ilshift__(self, other):
        newexpr = self._expr
        if newexpr:
            return newexpr << other
        else:
            return self._get_value() << other

    def __imatmul__(self, other):
        newexpr = self._expr
        if newexpr:
            return newexpr @ other
        else:
            return self._get_value() @ other

    def __imod__(self, other):
        newexpr = self._expr
        if newexpr:
            return newexpr % other
        else:
            return self._get_value() % other

    def __imul__(self, other):
        newexpr = self._expr
        if newexpr:
            return newexpr * other
        else:
            return self._get_value() * other

    def __ipow__(self, other):
        newexpr = self._expr
        if newexpr:
            return newexpr**other
        else:
            return self._get_value() ** other

    def __irshift__(self, other):
        newexpr = self._expr
        if newexpr:
            return newexpr >> other
        else:
            return self._get_value() >> other

    def __isub__(self, other):
        newexpr = self._expr
        if newexpr:
            return newexpr - other
        else:
            return self._get_value() - other

    def __itruediv__(self, other):
        newexpr = self._expr
        if newexpr:
            return newexpr / other
        else:
            return self._get_value() / other

    def __ixor__(self, other):
        newexpr = self._expr
        if newexpr:
            return newexpr ^ other
        else:
            return self._get_value() ^ other


class AttrRef(MutableRef):
    __slots__ = ("_owner", "_key", "_manager")

    def __init__(self, _owner, _key, _manager=None):
        object.__setattr__(self, "_owner", _owner)
        object.__setattr__(self, "_key", _key)
        object.__setattr__(self, "_manager", _manager)

    # Untested:

    # def __getstate__(self):
    #     return self._owner, self._key, self._manager

    # def __setstate__(self, state):
    #     object.__setattr__(self, "_owner", state[0])
    #     object.__setattr__(self, "_key", state[1])
    #     object.__setattr__(self, "_manager", state[2])

    def __hash__(self):
        if isinstance(self._owner, ARef):
            own = self._owner
        else:
            own = id(self._owner)
        return hash((own, self._key))

    def _get_value(self):
        owner = ARef._mk_value(self._owner)
        attr = ARef._mk_value(self._key)
        return getattr(owner, attr)

    def _set_value(self, value):
        owner = ARef._mk_value(self._owner)
        attr = ARef._mk_value(self._key)
        setattr(owner, attr, value)

    def _get_dependencies(self, out=None):
        if out is None:
            out = set()
        if isinstance(self._owner, ARef):
            self._owner._get_dependencies(out)
        if isinstance(self._key, ARef):
            self._key._get_dependencies(out)
        out.add(self)
        return out

    def __repr__(self):
        return f"{self._owner}.{self._key}"


class Ref(MutableRef):
    __slots__ = ("_owner", "_manager", "_label")

    def __init__(self, _owner, _manager=None, _label="_"):
        object.__setattr__(self, "_owner", _owner)
        object.__setattr__(self, "_manager", _manager)
        object.__setattr__(self, "_label", _label)

    # Untested:

    # def __getstate__(self):
    #     return self._owner, self._manager, self._label

    # def __setstate__(self, state):
    #     object.__setattr__(self, "_owner", state[0])
    #     object.__setattr__(self, "_manager", state[1])
    #     object.__setattr__(self, "_label", state[2])

    def __hash__(self):
        return hash(self._label)

    def __repr__(self):
        return self._label

    def _get_value(self):
        return ARef._mk_value(self._owner)


class ItemRef(MutableRef):
    __slots__ = ("_owner", "_key", "_manager")

    def __init__(self, _owner, __key, _manager=None):
        object.__setattr__(self, "_owner", _owner)
        object.__setattr__(self, "_key", __key)
        object.__setattr__(self, "_manager", _manager)

    # Untested:

    # def __getstate__(self):
    #     return self._owner, self._key, self._manager

    # def __setstate__(self, state):
    #     object.__setattr__(self, "_owner", state[0])
    #     object.__setattr__(self, "_key", state[1])
    #     object.__setattr__(self, "_manager", state[2])

    def __hash__(self):
        if isinstance(self._owner, ARef):
            own = self._owner
        else:
            own = id(self._owner)
        return hash((own, self._key))

    def _get_value(self):
        owner = ARef._mk_value(self._owner)
        item = ARef._mk_value(self._key)
        return owner[item]

    def _set_value(self, value):
        owner = ARef._mk_value(self._owner)
        item = ARef._mk_value(self._key)
        owner[item] = value

    def _get_dependencies(self, out=None):
        if out is None:
            out = set()
        if isinstance(self._owner, ARef):
            self._owner._get_dependencies(out)
        if isinstance(self._key, ARef):
            self._key._get_dependencies(out)
        out.add(self)
        return out

    def __repr__(self):
        return f"{self._owner}[{repr(self._key)}]"


class ItemDefaultRef(MutableRef):
    __slots__ = ("_owner", "_key", "_manager", "_default")

    def __init__(self, _owner, _key, _manager=None, _default=0):
        object.__setattr__(self, "_owner", _owner)
        object.__setattr__(self, "_key", _key)
        object.__setattr__(self, "_manager", _manager)
        object.__setattr__(self, "_default", _default)

    # Untested:

    # def __getstate__(self):
    #     return self._owner, self._key, self._manager, self._default

    # def __setstate__(self, state):
    #     object.__setattr__(self, "_owner", state[0])
    #     object.__setattr__(self, "_key", state[1])
    #     object.__setattr__(self, "_manager", state[2])
    #     object.__setattr__(self, "_default", state[3])

    def __hash__(self):
        if isinstance(self._owner, ARef):
            own = self._owner
        else:
            own = id(self._owner)
        return hash((own, self._key))

    def _get_value(self):
        owner = ARef._mk_value(self._owner)
        item = ARef._mk_value(self._key)
        return owner[item]

    def _set_value(self, value):
        owner = ARef._mk_value(self._owner)
        item = ARef._mk_value(self._key)
        owner[item] = value

    def _get_dependencies(self, out=None):
        if out is None:
            out = set()
        if isinstance(self._owner, ARef):
            self._owner._get_dependencies(out)
        if isinstance(self._key, ARef):
            self._key._get_dependencies(out)
        out.add(self)
        return out

    def __repr__(self):
        return f"{self._owner}[{repr(self._key)}]"


class ObjectAttrRef(Ref):

    # Untested:

    # def __getstate__(self):
    #     return self._manager

    # def __setstate__(self, state):
    #     object.__setattr__(self, "_manager", state)

    def __getattr__(self, attr):
        return ItemDefaultRef(self, attr, self._manager)

    def __setattr__(self, attr, value):
        ref = ItemDefaultRef(self, attr, self._manager)
        self._manager.set_value(ref, value)


@dataclass(frozen=True)
class BinOpRef(ARef):
    _a: object
    _b: object

    # Untested:

    # def __getstate__(self):
    #     return self._a, self._b

    # def __setstate__(self, state):
    #     object.__setattr__(self, "_a", state[0])
    #     object.__setattr__(self, "_b", state[1])

    def _get_value(self):
        a = ARef._mk_value(self._a)
        b = ARef._mk_value(self._b)
        return self._op(a, b)

    def _get_dependencies(self, out=None):
        a = self._a
        b = self._b
        if out is None:
            out = set()
        if isinstance(a, ARef):
            a._get_dependencies(out)
        if isinstance(b, ARef):
            b._get_dependencies(out)
        return out

    def __repr__(self):
        return f"({self._a}{self._st}{self._b})"


@dataclass(frozen=True)
class UnOpRef(ARef):
    _a: object

    # Untested:

    # def __getstate__(self):
    #     return self._a

    # def __setstate__(self, state):
    #     object.__setattr__(self, "_a", state)

    def _get_value(self):
        a = ARef._mk_value(self._a)
        return self._op(a)

    def _get_dependencies(self, out=None):
        a = self._a
        if out is None:
            out = set()
        if isinstance(a, ARef):
            a._get_dependencies(out)
        return out

    def __repr__(self):
        return f"({self._st}{self._a})"


@dataclass(frozen=True)
class BuiltinRef(ARef):
    _a: object

    # Untested:

    # def __getstate__(self):
    #     return self._a

    # def __setstate__(self, state):
    #     object.__setattr__(self, "_a", state)

    def _get_value(self):
        a = ARef._mk_value(self._a)
        return self._op(a)

    def _get_dependencies(self, out=None):
        a = self._a
        if out is None:
            out = set()
        if isinstance(a, ARef):
            a._get_dependencies(out)
        return out

    def __repr__(self):
        return f"{self._st}({self._a})"


@dataclass(frozen=True)
class CallRef(ARef):
    _func: object
    _args: tuple
    _kwargs: tuple

    # Untested:

    # def __getstate__(self):
    #     return self._func, self._args, self._kwargs

    # def __setstate__(self, state):
    #     object.__setattr__(self, "_func", state[0])
    #     object.__setattr__(self, "_args", state[1])
    #     object.__setattr__(self, "_kwargs", state[2])

    def _get_value(self):
        func = ARef._mk_value(self._func)
        args = [ARef._mk_value(a) for a in self._args]
        kwargs = {n: ARef._mk_value(v) for n, v in self._kwargs}
        return func(*args, **kwargs)

    def _get_dependencies(self, out=None):
        if out is None:
            out = set()
        if isinstance(self._func, ARef):
            self._func._get_dependencies(out)
        for arg in self._args:
            if isinstance(arg, ARef):
                arg._get_dependencies(out)
        for name, arg in self._kwargs:
            if isinstance(arg, ARef):
                arg._get_dependencies(out)
        return out

    def __repr__(self):
        args = []
        for aa in self._args:
            args.append(repr(aa))
        for k, v in self._kwargs:
            args.append(f"{k}={v}")
        args = ",".join(args)
        if isinstance(self._func, ARef):
            fname = repr(self._func)
        else:
            fname = self._func.__name__
        return f"{fname}({args})"


gbl = globals()
for st, op in _binops.items():
    fn = op.__name__.replace("_", "")
    cn = f"{fn.capitalize()}Ref"
    mn = f"__{fn}__"
    gbl[cn] = type(cn, (BinOpRef,), {"_op": op, "_st": st})

for st, op in _unops.items():
    fn = op.__name__.replace("_", "")
    cn = f"{fn.capitalize()}Ref"
    if cn in gbl:
        raise ValueError
    mn = f"__{fn}__"
    gbl[cn] = type(cn, (UnOpRef,), {"_op": op, "_st": st})

for st, op in _builtins.items():
    fn = op.__name__.replace("_", "")
    cn = f"B{fn.capitalize()}Ref"
    if cn in gbl:
        raise ValueError
    mn = f"__{fn}__"
    gbl[cn] = type(cn, (BuiltinRef,), {"_op": op, "_st": st})



import numpy as np
from ..general import _print

from numpy.linalg import lstsq

class JacobianSolver:

    def __init__(self, func, limits, n_steps_max=20, tol=1e-20, n_bisections=3,
                 min_step=1e-20, verbose=False):
        self.func = func
        self.limits = limits
        self.n_steps_max = n_steps_max
        self.tol = tol
        self.n_bisections = n_bisections
        self.min_step = min_step

        self._penalty_best = None
        self._xbest = None
        self.verbose = verbose

    def _eval(self, x):
        y = self.func(x)
        penalty = np.sqrt(np.dot(y, y))
        if self.verbose:
            _print(f"penalty: {penalty}")
        if penalty < self._penalty_best:
            if self._penalty_best - penalty > 1e-20: #????????????
                self._step_best = self._step
            self._penalty_best = penalty
            self._xbest = x.copy()
            if self.verbose:
                _print(f"new best: {self._penalty_best}")
        return y, penalty

    def solve(self, x0):

        myf = self.func

        x0 = np.array(x0, dtype=float)
        x = x0.copy()

        self._xbest = x0.copy()
        self._penalty_best = 1e200
        ncalls = 0
        info = {}
        mask_for_next_step = np.ones(len(x0), dtype=bool)
        for step in range(self.n_steps_max):

            self._step = step
            # test penalty
            y, penalty = self._eval(x) # will need to handle mask
            ncalls += 1
            if penalty < self.tol:
                if self.verbose:
                    _print("Jacobian tolerance met")
                break
            # Equation search
            jac = myf.get_jacobian(x) # will need to handle mask
            ncalls += len(x)

            # lstsq using only the the variables that were not at the limit
            # in the previous step
            xstep = np.zeros(len(x))

            xstep[mask_for_next_step] = lstsq(jac[:, mask_for_next_step], y, rcond=None)[0]  # newton step
            mask_for_next_step[:] = True
            self._last_xstep = xstep.copy()

            newpen = penalty * 2
            alpha = -1

            while newpen > penalty:  # bisec search
                if alpha > self.n_bisections:
                    break
                alpha += 1
                l = 2.0**-alpha
                if self.verbose:
                    _print(f"\n--> step {step} alpha {alpha}\n")

                this_xstep = l * xstep
                mask_hit_limit = np.zeros(len(x), dtype=bool)
                for ii in range(len(x)):
                    if x[ii] - this_xstep[ii] < self.limits[ii][0]:
                        this_xstep[ii] = 0
                        mask_hit_limit[ii] = True
                    elif x[ii] - this_xstep[ii] > self.limits[ii][1]:
                        this_xstep[ii] = 0
                        mask_hit_limit[ii] = True

                y, newpen = self._eval(x - this_xstep) # will need to handle mask

                ncalls += 1
            x -= this_xstep  # update solution
            mask_for_next_step = ~mask_hit_limit

            if self.verbose:
                _print(f"step {step} step_best {self._step_best} {this_xstep}")
            if np.sqrt(np.dot(this_xstep, this_xstep)) < self.min_step:
                if self.verbose:
                    _print("No progress, stopping")
                break
        else:
            if self.verbose:
                _print("Max steps reached")

        return self._xbest



# copyright ############################### #
# This file is part of the Xdeps Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

from dataclasses import dataclass, field
from collections import defaultdict
import logging
from copy import deepcopy

from .refs import ARef, Ref, ObjectAttrRef
from .refs import AttrRef, CallRef, ItemRef
from .utils import os_display_png, mpl_display_png, ipy_display_png
from .utils import AttrDict
from .sorting import toposort

logger = logging.getLogger(__name__)


def dct_merge(dct1, dct2):
    return {**dct1, **dct2}


def _check_root_owner(t, ref):
    if hasattr(t, "_owner"):
        if t._owner is ref:
            return True
        else:
            return _check_root_owner(t._owner, ref)
    else:
        return False


class FuncWrapper:
    def __init__(self, func):
        self.func = func

    def __call__(self, *args, **kwargs):
        return CallRef(self.func, args, tuple(kwargs.items()))


class Task:
    taskid: object
    targets: set
    dependencies: set

    def run(self):
        raise NotImplemented


class GenericTask(Task):
    taskid: object
    action: object
    targets: set
    dependencies: set

    def __repr__(self):
        return f"<Task {self.taskid}:{self.dependencies}=>{self.targets}>"

    def run(self, *args):
        return self.action(*args)


class ExprTask(Task):
    def __init__(self, target, expr):
        self.taskid = target
        self.targets = target._get_dependencies()
        self.dependencies = expr._get_dependencies()
        self.expr = expr

    def __repr__(self):
        return f"{self.taskid} = {self.expr}"

    def run(self):
        value = self.expr._get_value()
        self.taskid._set_value(value)

    def info(self):
        print(f"#  {self.taskid}._expr._get_dependencies()")
        for pp in self.expr._get_dependencies():
            print(f"   {pp} = {pp._get_value()}")
        print()


class InheritanceTask(Task):
    def __init__(self, children, parents):
        self.taskid = children
        self.targets = set([children])
        self.dependencies = set(parents)

    def __repr__(self):
        return f"{self.taskid} <- {self.parents}"

    def run(self, event):
        key, value, isattr = event
        for target in self.targets:
            if isattr:
                getattr(target, key)._set_value(value)
            else:
                target[key]._set_value(value)


class DepEnv:
    __slots__ = ("_data", "_")

    def __init__(self, data, ref):
        object.__setattr__(self, "_data", data)
        object.__setattr__(self, "_", ref)

    def __getattr__(self, key):
        return getattr(self._data, key)

    def __getitem__(self, key):
        return self._data[key]

    def __setattr__(self, key, value):
        self._[key] = value

    def __setitem__(self, key, value):
        self._[key] = value

    def _eval(self, expr):
        return self._._eval(expr)


class Manager:
    """
    Value dependency manager:

    tasks: taskid -> task
    rdeps: ref -> set of all refs that depends on `ref`
    rtasks: taskid -> set all tasks whose dependencies are affected by taskid
    deptasks: ref -> all tasks that has ref as dependency
    tartasks: ref -> all tasks that has ref as target
    containers: label -> controlled container
    """

    def __init__(self):
        self.tasks = {}
        self.containers = {}
        self.rdeps = defaultdict(list)
        self.rtasks = defaultdict(list)
        self.deptasks = defaultdict(list)
        self.tartasks = defaultdict(list)

    def ref(self, container=None, label="_"):
        """Return a ref to an instance (or dict) associated to a label.

        Label must be unique.
        """
        if container is None:
            container = AttrDict()
        objref = Ref(container, self, label)
        assert label not in self.containers
        self.containers[label] = objref
        return objref

    def set_value(self, ref, value):
        """Set a value pointed by a ref and execute all tasks that depends on ref.

        If the value is a Ref, create a new task from the ref.
        """
        logger.info("set_value %s %s", ref, value)
        if ref in self.tasks:
            self.unregister(ref)
        if isinstance(value, ARef):  # value is an expression
            self.register(ExprTask(ref, value))
            value = value._get_value()  # to be updated
        ref._set_value(value)
        self._run_tasks(self.find_tasks(ref._get_dependencies()))

    def _run_tasks(self, tasks):
        for task in tasks:
            logger.info("Run %s", task)
            task.run()

    def register(self, task):
        """Register a new task identified by taskid"""
        # logger.info("register %s",taskid)
        taskid = task.taskid
        self.tasks[taskid] = task
        for dep in task.dependencies:
            # logger.info("%s have an impact on %s",dep,task.targets)
            self.rdeps[dep].extend(task.targets)
            # logger.info("%s is used by T:%s",dep,taskid)
            self.deptasks[dep].append(taskid)
            for deptask in self.tartasks[dep]:
                # logger.info("%s modifies deps of T:%s",deptask,taskid)
                self.rtasks[deptask].append(taskid)
        for tar in task.targets:
            # logger.info("%s is modified by T:%s",tar,taskid)
            self.tartasks[tar].append(taskid)
            for deptask in self.deptasks[tar]:
                # logger.info("T:%s modifies deps of T:%s",taskid,deptask)
                self.rtasks[taskid].append(deptask)

    def unregister(self, taskid):
        """Unregister the task identified by taskid"""
        task = self.tasks[taskid]
        for dep in task.dependencies:
            for target in task.targets:
                if target in self.rdeps[dep]:
                    self.rdeps[dep].remove(target)
            if dep in self.rtasks[dep]:
                self.rtasks[dep].remove(taskid)
            if taskid in self.deptasks[dep]:
                self.deptasks[dep].remove(taskid)
        for tar in task.targets:
            self.tartasks[tar].remove(taskid)
            for deptask in self.deptasks[tar]:
                if taskid in self.rtasks[deptask]:
                    self.rtasks[taskid].remove(deptask)
        if taskid in self.rtasks:
            del self.rtasks[taskid]
        del self.tasks[taskid]

    def find_deps(self, start_set):
        """Find all refs that depends on ref in start_seps"""
        assert type(start_set) in (list, tuple, set)
        deps = toposort(self.rdeps, start_set)
        return deps

    def find_taskids_from_tasks(self, start_tasks=None):
        """Find all taskids whose dependencies are affected by the tasks in start_tasks"""
        if start_tasks is None:
            start_tasks = self.rtasks
        tasks = toposort(self.rtasks, start_tasks)
        return tasks

    def find_taskids(self, start_deps=None):
        """Find all taskids that depend on the refs in start_deps"""
        if start_deps is None:
            start_deps = self.rdeps
        start_tasks = set()
        for dep in start_deps:
            start_tasks.update(self.deptasks[dep])
        tasks = toposort(self.rtasks, start_tasks)
        return tasks

    def find_tasks(self, start_deps=None):
        """Find all tasks that depend on the refs in start_deps"""
        if start_deps is None:
            start_deps = self.rdeps
        return [self.tasks[taskid] for taskid in self.find_taskids(start_deps)]

    def iter_expr_tasks_owner(self, ref):
        """Return all ExprTask defintions that write registered container"""
        for t in self.find_tasks():
            # TODO check for all targets or limit to ExprTask
            if _check_root_owner(t.taskid, ref):
                yield str(t.taskid), str(t.expr)

    def copy_expr_from(self, mgr, name, bindings=None):
        """
        Copy expression from another manager

        name: one of toplevel container in mgr
        bindings: dictionary mapping old container names into new container refs
        """
        ref = mgr.containers[name]
        if bindings is None:
            cmbdct = self.containers
        else:
            cmbdct = dct_merge(self.containers, bindings)
        self.load(mgr.iter_expr_tasks_owner(ref), cmbdct)

    def mk_fun(self, name, **kwargs):
        """Write a python function that executes a set of tasks in order of dependencies:
        name: name of the functions
        kwards:
            the keys are used to defined the argument name of the functions
            the values are the refs that will be set
        """
        varlist, start = list(zip(*kwargs.items()))
        tasks = self.find_tasks(start)
        fdef = [f"def {name}({','.join(varlist)}):"]
        for vname, vref in kwargs.items():
            fdef.append(f"  {vref} = {vname}")
        for tt in tasks:
            fdef.append(f"  {tt}")
        fdef = "\n".join(fdef)
        return fdef

    def gen_fun(self, name, **kwargs):
        """Return a python function that executes a set of tasks in order of dependencies:
        name: name of the functions
        kwards:
            the keys are used to defined the argument name of the functions
            the values are the refs that will be set
        """
        fdef = self.mk_fun(name, **kwargs)
        gbl = {}
        lcl = {}
        gbl.update((k, r._owner) for k, r in self.containers.items())
        exec(fdef, gbl, lcl)
        return lcl[name]

    def plot_deps(self, start=None, backend="ipy"):
        """Plot a graph of task and target dependencies from start.

        Possible backend:
            mpl: generate a figure in matplotlib
            os: generate a file /tmp/out.png and use `display` to show it
            ipy: use Ipython facility for Jupyter notebooks
        """
        from pydot import Dot, Node, Edge

        if start is None:
            start = list(self.rdeps)
        pdot = Dot("g", graph_type="digraph", rankdir="LR")
        for task in self.find_tasks(start):
            tn = Node(" " + str(task.taskid), shape="circle")
            pdot.add_node(tn)
            for tt in task.targets:
                pdot.add_node(Node(str(tt), shape="square"))
                pdot.add_edge(Edge(tn, str(tt), color="blue"))
            for tt in task.dependencies:
                pdot.add_node(Node(str(tt), shape="square"))
                pdot.add_edge(Edge(str(tt), tn, color="blue"))
        png = pdot.create_png()
        if backend == "mpl":
            mpl_display_png(png)
        elif backend == "os":
            os_display_png(png)
        elif backend == "ipy":
            ipy_display_png(png)
        return pdot

    def plot_tasks(self, start=None, backend="ipy"):
        """Plot a graph of task dependencies

        Possible backend:
            mpl: generate a figure in matplotlib
            os: generate a file /tmp/out.png and use `display` to show it
            ipy: use Ipython facility for Jupyter notebooks
        """
        from pydot import Dot, Node, Edge

        if start is None:
            start = list(self.rdeps)
        pdot = Dot("g", graph_type="digraph", rankdir="LR")
        for task in self.find_tasks(start):
            tn = Node(str(task.taskid), shape="circle")
            pdot.add_node(tn)
            for dep in task.dependencies:
                for tt in self.tartasks[dep]:
                    pdot.add_edge(Edge(str(tt), tn, color="blue"))
        png = pdot.create_png()
        if backend == "mpl":
            mpl_display_png(png)
        elif backend == "os":
            os_display_png(png)
        elif backend == "ipy":
            ipy_display_png(png)
        return pdot

    def dump(self):
        """Dump in json all ExprTask defined in the manager"""
        data = [
            (str(tt.taskid), str(tt.expr))
            # for t in self.find_tasks(self.rdeps)
            for tt in self.tasks.values()
            if isinstance(tt, ExprTask)
        ]
        return data

    def load(self, dump, dct=None):
        """Reload the expressions in dump  using container in dct

        dump: list of (lhs,rhs) pairs
        dct: dictionary of named references of containers,
             self containers by default

        """
        if dct is None:
            dct = self.containers
        for lhs, rhs in dump:
            lhs = eval(lhs, {}, dct)
            rhs = eval(rhs, {}, dct)
            task = ExprTask(lhs, rhs)
            self.register(task)

    def newenv(self, label="_", data=None):
        "Experimental"
        if data is None:
            data = AttrDict()
        ref = self.ref(data, label=label)
        return DepEnv(data, ref)

    def refattr(self, container=None, label="_"):
        "Experimental"
        if container is None:
            container = AttrDict()
        objref = ObjectAttrRef(container, self, label)
        assert label not in self.containers
        self.containers[label] = objref
        return objref

    def cleanup(self):
        """
        Remove empty sets from dicts
        """
        for dct in self.rdeps, self.rtasks, self.deptasks, self.tartasks:
            for kk, ss in list(dct.items()):
                if len(ss) == 0:
                    del dct[kk]

    def copy(self):
        """
        Create a copy of in new manager
        """
        other = Manager()
        other.containers = deepcopy(self.containers)
        other.tasks = deepcopy(self.tasks)
        other.rdeps = deepcopy(self.rdeps)
        other.rtasks = deepcopy(self.rtasks)
        other.deptasks = deepcopy(self.deptasks)
        other.tartasks = deepcopy(self.tartasks)
        return other

    def rebuild(self):
        self.cleanup()
        other = Manager()
        other.containers.update(self.containers)
        for task in self.tasks.values():
            other.register(task)
        other.cleanup()
        return other

    def verify(self, dcts=("rdeps", "rtasks", "deptasks", "tartasks")):
        other = self.rebuild()
        for dct in dcts:
            odct = getattr(other, dct)
            sdct = getattr(self, dct)
            for kk, ss in list(sdct.items()):
                if set(ss) != set(odct[kk]):
                    print(f"{dct}[{kk}] not consistent")
                    print(f"{dct}[{kk}] self - check:", set(ss) - set(odct[kk]))
                    print(f"{dct}[{kk}] check - self:", set(odct[kk]) - set(ss))
                    # raise (ValueError(f"{self} is not consistent in {dct}[{kk}]"))



# copyright ############################### #
# This file is part of the Xdeps Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

from setuptools import setup, find_packages
from pathlib import Path

version_file = Path(__file__).parent / 'xdeps/_version.py'
dd = {}
with open(version_file.absolute(), 'r') as fp:
    exec(fp.read(), dd)
__version__ = dd['__version__']

setup(
        name='xdeps',
        version=__version__,
        description='Data dependency manager',
        long_description=("Data dependency manager\n"
                         "\nThis package is part of the Xsuite collection."),
        author='Riccardo De Maria',
        author_email='riccardo.de.maria@cern.ch',
        packages=find_packages(),
        install_requires=['lark'],
        url='https://xsuite.readthedocs.io/',
        license='Apache 2.0',
        download_url="https://pypi.python.org/pypi/xdeps",
        project_urls={
            "Bug Tracker": "https://github.com/xsuite/xsuite/issues",
            "Documentation": 'https://xsuite.readthedocs.io/',
            "Source Code": "https://github.com/xsuite/xdeps",
        },
        extras_require={
            'tests': ['pytest'],
        },
)
